#!/usr/bin/env python
# coding: utf-8
from __future__ import print_function

import os
import sys
import time
import socket
import struct

"""
ripcap.py: pcap parser except it's fast -- 250 MiB/s, 20K pkt/s
"""
__version__   = "1.3"
__date__      = "2020-05-12"
__author__    = "ed <irc.rizon.net>"
__url__       = "https://ocv.me/dev/?ripcap.py"
__credits__   = ["stackoverflow.com"]
__license__   = "MIT"
__copyright__ = 2020


# config
ALLOW_UNKNOWN_PCAPNG_BLOCKS = True
ALLOW_UNKNOWN_PCAPNG_BLOCKS = False


PY2 = sys.version_info[0] == 2
if PY2:
    from cStringIO import StringIO as BytesIO
else:
    from io import BytesIO as BytesIO


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def b2hex(data):
    if PY2:
        return " ".join(map(lambda b: format(ord(b), "02x"), data))
    else:
        if type(data) is str:
            return " ".join(map(lambda b: format(ord(b), "02x"), data))
        else:
            return " ".join(map(lambda b: format(b, "02x"), data))


class PositionTracker(object):
    def __init__(self, buf):
        self.f = buf
        self.pos = 0

    def read(self, n=-1):
        rv = self.f.read(n)
        self.pos += len(rv)
        return rv

    def tell(self):
        return self.pos


class Parser(object):
    def __init__(self, buf):
        if hasattr(buf, "read"):
            try:
                buf.tell()
                self.f = buf
                # eprint('[*] direct buffer access')
            except:
                self.f = PositionTracker(buf)
                # eprint('[*] indirect access using PositionTracker')
        else:
            # eprint('[*] indirect access using BytesIO')
            self.f = BytesIO(buf)

    def u64(self):
        buf = self.f.read(8)
        if len(buf) < 8:
            raise Exception("EOF")

        return struct.unpack("<Q", buf)[0]

    def ru64(self):
        buf = self.f.read(8)
        if len(buf) < 8:
            raise Exception("EOF")

        return struct.unpack(">Q", buf)[0]

    def s32(self):
        buf = self.f.read(4)
        if len(buf) < 4:
            raise Exception("EOF")

        return struct.unpack("<l", buf)[0]

    def u32(self):
        buf = self.f.read(4)
        if len(buf) < 4:
            raise Exception("EOF")

        return struct.unpack("<L", buf)[0]

    def ru32(self):
        buf = self.f.read(4)
        if len(buf) < 4:
            raise Exception("EOF")

        return struct.unpack(">L", buf)[0]

    def u16(self):
        buf = self.f.read(2)
        if len(buf) < 2:
            raise Exception("EOF")

        return struct.unpack("<H", buf)[0]

    def ru16(self):
        buf = self.f.read(2)
        if len(buf) < 2:
            raise Exception("EOF")

        return struct.unpack(">H", buf)[0]

    def u8(self):
        buf = self.f.read(1)
        if not buf:
            raise Exception("EOF")

        return struct.unpack("B", buf)[0]

    def drop(self, n_bytes):
        if len(self.f.read(n_bytes)) != n_bytes:
            raise Exception("EOF")

    def unk(self, expect):
        v = self.f.read(len(expect))
        if v != expect:
            raise Exception(
                "unexpected unknown:\n{0} got\n{1} expected".format(
                    b2hex(v), b2hex(expect)
                )
            )

    def str(self, maxlen):
        size = self.u32()
        if size > maxlen:
            raise Exception("string out of bounds: {0}".format(size))

        return self.f.read(size).decode("utf-8").rstrip("\0")

    def align(self, to):
        pos = self.f.tell()
        waste = 0
        while divmod(pos, to)[1] != 0:
            waste += 1
            pos += 1

        if waste != 0:
            self.f.read(waste)


class Ripcap(object):
    def __init__(self, fobj, status_interval=128):
        self.fobj = fobj
        self.status_interval = status_interval
        self.num_packets = 0
        self.num_bytes = 0
        self.last_num_packets = 0
        self.last_num_bytes = 0
        self.last_print = 0

    def handle_packet(self, pcap, buf):
        if self.status_interval and self.num_packets % self.status_interval == 0:
            now = time.time()
            if self.last_print != 0:
                td = now - self.last_print
                byte_sec = (self.num_bytes - self.last_num_bytes) / td
                pkt_sec = (self.num_packets - self.last_num_packets) / td
            else:
                byte_sec = 0
                pkt_sec = 0

            self.last_print = now
            self.last_num_bytes = self.num_bytes
            self.last_num_packets = self.num_packets

            m = "\033[1;30m[*] {:.3f}k packets, {:.3f} kB, {:.3f}k pkt/s, {:.3f} kB/s\033[0m".format(
                self.num_packets / 1000.0,
                self.num_bytes / 1000.0,
                pkt_sec / 1000.0,
                byte_sec / 1000.0,
            )
            eprint(m)

        p = Parser(buf)
        # read(6) destination mac
        # read(6) source mac
        p.drop(12)

        pkt_type = p.u16()
        if pkt_type not in [0x8, 0xdd86]:
            raise RuntimeError(
                "at packet offset 0x{:x} in pcap offset 0x{:x}, expected IPv4 (0x8) or IPv6 (0x86dd), got 0x{:x}".format(
                    pcap.f.tell(), p.f.tell(), pkt_type
                )
            )

        if pkt_type == 0x8:
            blk_pos = p.f.tell()
            try:
                # u8 0x40 (ipv4) + header length
                # u8 differentiated services field + explicit congestion notification
                p.drop(2)
                frame_sz = p.ru16()
                p.drop(2) # ru16 seqno
            except Exception as e:
                raise RuntimeError(
                    "at packet offset 0x{:x} in pcap offset 0x{:x}, got unexpected data".format(
                        pcap.f.tell(), p.f.tell()
                    )
                )

            try:
                p.unk(b"\x40")  # flags: dont fragment, no more fragments
                p.unk(b"\x00")  # fragment offset: 0
                p.u8()  # TTL
                fragmented = False
            except:
                fragmented = True
                eprint(
                    "[x] at pcap offset 0x{:x}, got fragmented packet, will discard".format(
                        p.f.tell()
                    )
                )
            
            protocol = p.u8()
            p.u16()  # checksum
            src_ip = socket.inet_ntoa(p.f.read(4))
            dst_ip = socket.inet_ntoa(p.f.read(4))
            
        else:
            # 4bit version + traffic class
            # 8bit traffic class
            # 5bit flow label
            p.drop(4)
            frame_sz = p.ru16()
            protocol = p.u8()
            p.drop(1) # hop limit
            src_ip = socket.inet_ntop(socket.AF_INET6, p.f.read(16))
            dst_ip = socket.inet_ntop(socket.AF_INET6, p.f.read(16))
            blk_pos = p.f.tell()
            fragmented = False  # ┐(ﾟ∀ﾟ)┌

        if protocol not in [0x11, 0x06]:
            raise RuntimeError(
                "at packet offset 0x{:x} in pcap offset 0x{:x}, expected UDP (0x11) or TCP (0x06), got 0x{:x}".format(
                    pcap.f.tell(), p.f.tell(), protocol
                )
            )

        if protocol == 0x11:
            src_port = p.ru16()
            dst_port = p.ru16()
            pkt_len = p.ru16() - 8
            p.u16()  # checksum
            len_chk = blk_pos + frame_sz - p.f.tell()
            if pkt_len != len_chk:
                raise RuntimeError(
                    "at packet offset 0x{:x} in pcap offset 0x{:x}, bad packet length: expected 0x{:x} bytes, got 0x{:x}".format(
                        pcap.f.tell(), p.f.tell(), len_chk, pkt_len
                    )
                )

        else:
            src_port = p.ru16()
            dst_port = p.ru16()
            # ru32 seq.no
            # ru32 ack.no
            p.drop(8)
            data_offset = int(p.u8() / 0x10) * 4
            p.drop(data_offset - (4 * 3 + 1))
            pkt_len = blk_pos + frame_sz - p.f.tell()

        buf = b''
        left = pkt_len
        while left > 0:
            t = p.f.read(left)
            if not t:
                raise RuntimeError(
                "at packet offset 0x{:x} in pcap offset 0x{:x}, partial packet: expected 0x{:x} bytes, got 0x{:x}".format(
                    pcap.f.tell(), p.f.tell(), pkt_len, len(buf)
                )
            )
            left -= len(t)
            buf += t
    
        self.num_packets += 1
        self.num_bytes += len(buf)
        if fragmented:
            return None
        else:
            return src_ip, src_port, dst_ip, dst_port, buf

    def parse_pcap(self):
        p = Parser(self.fobj)

        bom = p.u32()
        if bom == 0x0A0D0D0A:
            raise RuntimeError("\n\n    this is a pcapng file, not pcap\n")
        if bom != 0xA1B2C3D4:
            raise RuntimeError("unexpected endianness in pcap: {:08x}".format(bom))

        ver_maj = p.u16()
        ver_min = p.u16()
        if ver_maj != 2:
            # tested with 2.4
            raise RuntimeError(
                "unsupported pcap version: {0}.{1}".format(ver_maj, ver_min)
            )

        p.unk(b"\x00\x00\x00\x00")  # timezone offset in seconds; always GMT
        p.unk(b"\x00\x00\x00\x00")  # time accuracy; always 0
        
        # u32 max packet length (usually 65535, dontcare anyways)
        # u32 network type (ethernet = 1)
        p.drop(8)

        while True:
            try:
                ts_hi = p.u32()
            except:
                return

            ts_lo = p.u32()
            cap_len = p.u32()
            pkt_len = p.u32()
            if cap_len != pkt_len:
                raise RuntimeError(
                    "not impl: partial packet at 0x{:x}; {} {}".format(
                        p.f.tell(), cap_len, pkt_len
                    )
                )

            buf = p.f.read(pkt_len)
            ret = self.handle_packet(p, buf)
            if ret:
                yield [ts_hi + ts_lo / (1000.0 * 1000)] + list(ret)

    def parse_pcapng(self):
        p = Parser(self.fobj)
        while True:
            blk_pos = p.f.tell()
            try:
                blk_type = p.u32()
                blk_len = p.u32()
            except:
                break  # eof

            if blk_type == 0x0A0D0D0A:  # Section Header Block
                bom = p.u32()
                if bom != 0x1A2B3C4D:
                    raise RuntimeError("unexpected endianness in pcapng")

                ver_maj = p.u16()
                ver_min = p.u16()
                if ver_maj != 1 or ver_min != 0:
                    raise RuntimeError(
                        "unsupported pcapng version: {0}.{1}".format(ver_maj, ver_min)
                    )

            elif blk_type == 0x1:  # Interface Description Block
                pass

            elif blk_type == 0x6:  # Enhanced Packet block
                p.u32()  # interface id
                ts_hi = p.u32()
                ts_lo = p.u32()
                cap_len = p.u32()
                pkt_len = p.u32()
                if cap_len != pkt_len:
                    raise RuntimeError(
                        "not impl: partial packet at 0x{:x}".format(p.f.tell())
                    )

                buf = p.f.read(pkt_len)
                p.align(4)
                ret = self.handle_packet(p, buf)
                if ret:
                    yield [ts_hi + ts_lo / (1000.0 * 1000 * 1000)] + list(ret)

            elif blk_type == 0x3:  # Simple Packet block (UNTESTED)
                pkt_len = p.u32()
                buf = p.f.read(pkt_len)
                p.align(4)
                ret = self.handle_packet(p, buf)
                if ret:
                    yield [time.time()] + list(ret)

            elif blk_type == 0x5:  # Interface Statistics Block
                pass

            elif blk_type == 0xA1B2C3D4:
                raise RuntimeError("\n\n    this is a pcap file, not pcapng\n")

            else:
                if not ALLOW_UNKNOWN_PCAPNG_BLOCKS:
                    raise RuntimeError(
                        "unsupported pcapng block type: {:x}".format(blk_type)
                    )

            # final 4 bytes of each block is another repetition of blk_len,
            # use this as a sanity check
            p.drop((blk_len - (p.f.tell() - blk_pos)) - 4)
            v = p.u32()
            if v != blk_len:
                raise RuntimeError(
                    "corrupt pcapng; expected {:x}, got {:x}".format(blk_len, v)
                )


class Tcpdump(object):
    def __init__(self, iface, cap_filter):
        import subprocess as sp
        
        cmd = ['tcpdump', '-Unni', iface, '--immediate-mode', '-w', '-']
        if cap_filter:
            cmd.append(cap_filter)

        self.p = sp.Popen(cmd, stdout=sp.PIPE)
        self.read = self.p.stdout.read
    
    def stream(self):
        while True:
            buf = self.p.stdout.read(4096)
            if buf:
                yield buf
            else:
                if self.p.poll() is not None:
                    return


def main():
    import argparse

    ap = argparse.ArgumentParser()
    ap.add_argument('-ng', action='store_true', help='assume pcapng input')
    ap.add_argument('-v', action='store_true', help='print packet contents')
    ap.add_argument('-i', metavar='IFACE', type=str, help='interface to capture')
    ap.add_argument('-f', metavar='FILTER', type=str, help='tcpdump filter')
    ar = ap.parse_args()

    if not PY2:
        stdin = sys.stdin.buffer
    else:
        if sys.platform == "win32":
            import msvcrt

            msvcrt.setmode(sys.stdin.fileno(), os.O_BINARY)

        stdin = sys.stdin

    src = 'stdin'
    src_fobj = stdin
    if ar.i:
        src = 'tcpdump subprocess'
        src_fobj = Tcpdump(ar.i, ar.f)
    
    if ar.ng:
        eprint("[*] reading pcapng from", src)
    else:
        eprint("[*] reading pcap from", src, "(use -ng for pcapng input)")

    if not ar.i:
        eprint("[*] usage example:")
        eprint("    tcpdump -Unni lo --immediate-mode -w - | ./readpcap.py")
        eprint()

    p = Ripcap(src_fobj)
    pp = p.parse_pcapng if ar.ng else p.parse_pcap

    for ts, src_ip, src_port, dst_ip, dst_port, pkt in pp():
        if not ar.v:
            continue
        
        eprint(
            "[*] {:.3f}  {}/{}  {}/{}  len({})  [{}]".format(
                ts,
                src_ip,
                src_port,
                dst_ip,
                dst_port,
                len(pkt),
                pkt.decode("utf-8", "replace"),
            )
        )

    eprint("[+] done")


if __name__ == "__main__":
    main()


"""
minimal py3 example (save to same folder as this):

#!/usr/bin/env python3
from ripcap import Ripcap, Tcpdump, sys

# p = Ripcap(sys.stdin.buffer)
p = Ripcap(Tcpdump('lo', 'port 4321'))
for x in p.parse_pcap():
    print(x)



# generate packets
g() { while true; do date +%s.%N; sleep 0.1; done | stdbuf -o0 tr -d '\n'; }

# udp server/client:
socat udp4-recv:4321,reuseaddr -
g | socat - udp4-sendto:127.0.0.1:4321

# tcp server/client:
socat tcp-listen:4321,fork,reuseaddr -
g | socat - tcp-connect:127.0.0.1:4321

# tcp6:
socat tcp6-listen:4321,fork,reuseaddr -
g | socat - tcp-connect:'[::1]':4321

# udp6:
socat udp6-listen:4321,fork,reuseaddr -
g | socat - udp6-connect:'[::1]':4321

"""
